@@ -13,8 +13,8 @@ module Yesod.Core.Class.Handler
 import Yesod.Core.Types
 import Data.Monoid (mempty)
 import Control.Monad (liftM)
-import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Monad.Trans.Resource (MonadResource, MonadResourceBase, ExceptionT (..))
+import Control.Monad.IO.Class (liftIO)
+import Control.Monad.Trans.Resource (MonadResource, MonadResourceBase)
 import Control.Monad.Trans.Class (lift)
 import Data.Monoid (Monoid)
 import Data.Conduit.Internal (Pipe, ConduitM)
@@ -61,9 +61,6 @@ GOX(Monoid w, RWST r w s)
 GOX(Monoid w, Strict.RWST r w s)
 GO(Strict.StateT s)
 GOX(Monoid w, Strict.WriterT w)
-#if !MIN_VERSION_resourcet(1,1,0)
-GO(ExceptionT)
-#endif
 GO(Pipe l i o u)
 GO(ConduitM i o)
 #undef GO
@@ -87,9 +84,6 @@ GOX(Monoid w, RWST r w s)
 GOX(Monoid w, Strict.RWST r w s)
 GO(Strict.StateT s)
 GOX(Monoid w, Strict.WriterT w)
-#if !MIN_VERSION_resourcet(1,1,0)
-GO(ExceptionT)
-#endif
 GO(Pipe l i o u)
 GO(ConduitM i o)
 #undef GO
