@@ -52,6 +52,9 @@ module Yesod.Core.Handler
     , lookupCookie
     , lookupFile
     , lookupHeader
+      -- **** Lookup authentication data
+    , lookupBasicAuth
+    , lookupBearerAuth
       -- **** Multi-lookup
     , lookupGetParams
     , lookupPostParams
@@ -92,12 +95,9 @@ module Yesod.Core.Handler
     , sendResponseCreated
     , sendWaiResponse
     , sendWaiApplication
-#if MIN_VERSION_wai(2, 1, 0)
     , sendRawResponse
-#endif
-#if MIN_VERSION_wai(3, 0, 0)
     , sendRawResponseNoConduit
-#endif
+    , notModified
       -- * Different representations
       -- $representations
     , selectRep
@@ -116,6 +116,7 @@ module Yesod.Core.Handler
     , neverExpires
     , alreadyExpired
     , expiresAt
+    , setEtag
       -- * Session
     , SessionMap
     , lookupSession
@@ -139,6 +140,7 @@ module Yesod.Core.Handler
       -- ** Hamlet
     , hamletToRepHtml
     , giveUrlRenderer
+    , withUrlRenderer
       -- ** Misc
     , newIdent
       -- * Lifting
@@ -148,6 +150,7 @@ module Yesod.Core.Handler
     , getMessageRender
       -- * Per-request caching
     , cached
+    , cachedBy
     ) where
 
 import           Data.Time                     (UTCTime, addUTCTime,
@@ -166,6 +169,8 @@ import           Control.Monad.IO.Class        (MonadIO, liftIO)
 
 import qualified Network.HTTP.Types            as H
 import qualified Network.Wai                   as W
+import           Network.Wai.Middleware.HttpAuth
+    ( extractBasicAuth, extractBearerAuth )
 import Control.Monad.Trans.Class (lift)
 
 import qualified Data.Text                     as T
@@ -190,10 +195,9 @@ import           Yesod.Core.Content            (ToTypedContent (..), simpleConte
 import           Yesod.Core.Internal.Util      (formatRFC1123)
 import           Text.Blaze.Html               (preEscapedToMarkup, toHtml)
 
-import           Data.Dynamic                  (fromDynamic, toDyn)
 import qualified Data.IORef.Lifted             as I
 import           Data.Maybe                    (listToMaybe, mapMaybe)
-import           Data.Typeable                 (Typeable, typeOf)
+import           Data.Typeable                 (Typeable)
 import           Web.PathPieces                (PathPiece(..))
 import           Yesod.Core.Class.Handler
 import           Yesod.Core.Types
@@ -205,22 +209,14 @@ import Data.CaseInsensitive (CI)
 import qualified Data.Conduit.List as CL
 import Control.Monad (unless)
 import           Control.Monad.Trans.Resource  (MonadResource, InternalState, runResourceT, withInternalState, getInternalState, liftResourceT, resourceForkIO
-#if MIN_VERSION_wai(2, 0, 0)
-#else
-              , ResourceT
-#endif
               )
-#if MIN_VERSION_wai(2, 0, 0)
 import qualified System.PosixCompat.Files as PC
-#endif
-#if MIN_VERSION_wai(2, 1, 0)
 import Control.Monad.Trans.Control (control, MonadBaseControl)
-#endif
 import Data.Conduit (Source, transPipe, Flush (Flush), yield, Producer
-#if MIN_VERSION_wai(2, 1, 0)
                     , Sink
-#endif
                    )
+import qualified Yesod.Core.TypeCache as Cache
+import qualified Data.Word8 as W8
 
 get :: MonadHandler m => m GHState
 get = liftHandlerT $ HandlerT $ I.readIORef . handlerState
@@ -256,41 +252,22 @@ runRequestBody = do
         Just rbc -> return rbc
         Nothing -> do
             rr <- waiRequest
-#if MIN_VERSION_wai_extra(2, 0, 1)
             internalState <- liftResourceT getInternalState
             rbc <- liftIO $ rbHelper upload rr internalState
-#elif MIN_VERSION_wai(2, 0, 0)
-            rbc <- liftIO $ rbHelper upload rr
-#else
-            rbc <- liftResourceT $ rbHelper upload rr
-#endif
             put x { ghsRBC = Just rbc }
             return rbc
 
-#if MIN_VERSION_wai(2, 0, 0)
 rbHelper :: FileUpload -> W.Request -> InternalState -> IO RequestBodyContents
 rbHelper upload req internalState =
-#else
-rbHelper :: FileUpload -> W.Request -> ResourceT IO RequestBodyContents
-rbHelper upload req =
-#endif
     case upload of
         FileUploadMemory s -> rbHelper' s mkFileInfoLBS req
-#if MIN_VERSION_wai_extra(2, 0, 1)
         FileUploadDisk s -> rbHelper' (s internalState) mkFileInfoFile req
-#else
-        FileUploadDisk s -> rbHelper' s mkFileInfoFile req
-#endif
         FileUploadSource s -> rbHelper' s mkFileInfoSource req
 
 rbHelper' :: NWP.BackEnd x
           -> (Text -> Text -> x -> FileInfo)
           -> W.Request
-#if MIN_VERSION_wai(2, 0, 0)
           -> IO ([(Text, Text)], [(Text, FileInfo)])
-#else
-          -> ResourceT IO ([(Text, Text)], [(Text, FileInfo)])
-#endif
 rbHelper' backend mkFI req =
     (map fix1 *** mapMaybe fix2) <$> (NWP.parseRequestBody backend req)
   where
@@ -309,7 +286,7 @@ rbHelper' backend mkFI req =
 askHandlerEnv :: MonadHandler m => m (RunHandlerEnv (HandlerSite m))
 askHandlerEnv = liftHandlerT $ HandlerT $ return . handlerEnv
 
--- | Get the master site appliation argument.
+-- | Get the master site application argument.
 getYesod :: MonadHandler m => m (HandlerSite m)
 getYesod = rheSite `liftM` askHandlerEnv
 
@@ -374,11 +351,7 @@ handlerToIO =
           where
             oldReq    = handlerRequest oldHandlerData
             oldWaiReq = reqWaiRequest oldReq
-#if MIN_VERSION_wai(3, 0, 0)
             newWaiReq = oldWaiReq { W.requestBody = return mempty
-#else
-            newWaiReq = oldWaiReq { W.requestBody = mempty
-#endif
                                   , W.requestBodyLength = W.KnownLength 0
                                   }
         oldEnv = handlerEnv oldHandlerData
@@ -387,6 +360,7 @@ handlerToIO =
       return $ oldState { ghsRBC = Nothing
                         , ghsIdent = 1
                         , ghsCache = mempty
+                        , ghsCacheBy = mempty
                         , ghsHeaders = mempty }
 
     -- xx From this point onwards, no references to oldHandlerData xx
@@ -550,16 +524,12 @@ sendFilePart :: MonadHandler m
              -> Integer -- ^ count
              -> m a
 sendFilePart ct fp off count = do
-#if MIN_VERSION_wai(2, 0, 0)
     fs <- liftIO $ PC.getFileStatus fp
     handlerError $ HCSendFile ct fp $ Just W.FilePart
         { W.filePartOffset = off
         , W.filePartByteCount = count
         , W.filePartFileSize = fromIntegral $ PC.fileSize fs
         }
-#else
-    handlerError $ HCSendFile ct fp $ Just $ W.FilePart off count
-#endif
 
 -- | Bypass remaining handler code and output the given content with a 200
 -- status code.
@@ -592,7 +562,6 @@ sendWaiResponse = handlerError . HCWai
 sendWaiApplication :: MonadHandler m => W.Application -> m b
 sendWaiApplication = handlerError . HCWaiApp
 
-#if MIN_VERSION_wai(3, 0, 0)
 -- | Send a raw response without conduit. This is used for cases such as
 -- WebSockets. Requires WAI 3.0 or later, and a web server which supports raw
 -- responses (e.g., Warp).
@@ -603,14 +572,12 @@ sendRawResponseNoConduit
     => (IO S8.ByteString -> (S8.ByteString -> IO ()) -> m ())
     -> m a
 sendRawResponseNoConduit raw = control $ \runInIO ->
-    runInIO $ sendWaiResponse $ flip W.responseRaw fallback
+    liftIO $ throwIO $ HCWai $ flip W.responseRaw fallback
     $ \src sink -> runInIO (raw src sink) >> return ()
   where
     fallback = W.responseLBS H.status500 [("Content-Type", "text/plain")]
         "sendRawResponse: backend does not support raw responses"
-#endif
 
-#if MIN_VERSION_wai(2, 1, 0)
 -- | Send a raw response. This is used for cases such as WebSockets. Requires
 -- WAI 2.1 or later, and a web server which supports raw responses (e.g.,
 -- Warp).
@@ -619,9 +586,8 @@ sendRawResponseNoConduit raw = control $ \runInIO ->
 sendRawResponse :: (MonadHandler m, MonadBaseControl IO m)
                 => (Source IO S8.ByteString -> Sink S8.ByteString IO () -> m ())
                 -> m a
-#if MIN_VERSION_wai(3, 0, 0)
 sendRawResponse raw = control $ \runInIO ->
-    runInIO $ sendWaiResponse $ flip W.responseRaw fallback
+    liftIO $ throwIO $ HCWai $ flip W.responseRaw fallback
     $ \src sink -> runInIO (raw (src' src) (CL.mapM_ sink)) >> return ()
   where
     fallback = W.responseLBS H.status500 [("Content-Type", "text/plain")]
@@ -631,15 +597,13 @@ sendRawResponse raw = control $ \runInIO ->
         unless (S.null bs) $ do
             yield bs
             src' src
-#else
-sendRawResponse raw = control $ \runInIO ->
-    runInIO $ sendWaiResponse $ flip W.responseRaw fallback
-    $ \src sink -> runInIO (raw src sink) >> return ()
-  where
-    fallback = W.responseLBS H.status500 [("Content-Type", "text/plain")]
-        "sendRawResponse: backend does not support raw responses"
-#endif
-#endif
+
+-- | Send a 304 not modified response immediately. This is a short-circuiting
+-- action.
+--
+-- Since 1.4.4
+notModified :: MonadHandler m => m a
+notModified = sendWaiResponse $ W.responseBuilder H.status304 [] mempty
 
 -- | Return a 404 not found page. Also denotes no handler available.
 notFound :: MonadHandler m => m a
@@ -735,7 +699,7 @@ cacheSeconds i = setHeader "Cache-Control" $ T.concat
 -- is never (realistically) expired.
 neverExpires :: MonadHandler m => m ()
 neverExpires = do
-    setHeader "Expires" "Thu, 31 Dec 2037 23:55:55 GMT"
+    askHandlerEnv >>= liftIO . rheGetMaxExpires >>= setHeader "Expires"
     cacheSeconds oneYear
   where
     oneYear :: Int
@@ -750,6 +714,36 @@ alreadyExpired = setHeader "Expires" "Thu, 01 Jan 1970 05:05:05 GMT"
 expiresAt :: MonadHandler m => UTCTime -> m ()
 expiresAt = setHeader "Expires" . formatRFC1123
 
+-- | Check the if-none-match header and, if it matches the given value, return
+-- a 304 not modified response. Otherwise, set the etag header to the given
+-- value.
+--
+-- Note that it is the responsibility of the caller to ensure that the provided
+-- value is a value etag value, no sanity checking is performed by this
+-- function.
+--
+-- Since 1.4.4
+setEtag :: MonadHandler m => Text -> m ()
+setEtag etag = do
+    mmatch <- lookupHeader "if-none-match"
+    let matches = maybe [] parseMatch mmatch
+    if encodeUtf8 etag `elem` matches
+        then notModified
+        else addHeader "etag" $ T.concat ["\"", etag, "\""]
+
+-- | Parse an if-none-match field according to the spec. Does not parsing on
+-- weak matches, which are not supported by setEtag.
+parseMatch :: S.ByteString -> [S.ByteString]
+parseMatch =
+    map clean . S.split W8._comma
+  where
+    clean = stripQuotes . fst . S.spanEnd W8.isSpace . S.dropWhile W8.isSpace
+
+    stripQuotes bs
+        | S.length bs >= 2 && S.head bs == W8._quotedbl && S.last bs == W8._quotedbl
+            = S.init $ S.tail bs
+        | otherwise = bs
+
 -- | Set a variable in the user's session.
 --
 -- The session is handled by the clientsession package: it sets an encrypted
@@ -854,7 +848,7 @@ redirectToPost :: (MonadHandler m, RedirectUrl (HandlerSite m) url)
                -> m a
 redirectToPost url = do
     urlText <- toTextUrl url
-    giveUrlRenderer [hamlet|
+    withUrlRenderer [hamlet|
 $newline never
 $doctype 5
 
@@ -870,17 +864,26 @@ $doctype 5
 
 -- | Wraps the 'Content' generated by 'hamletToContent' in a 'RepHtml'.
 hamletToRepHtml :: MonadHandler m => HtmlUrl (Route (HandlerSite m)) -> m Html
-hamletToRepHtml = giveUrlRenderer
-{-# DEPRECATED hamletToRepHtml "Use giveUrlRenderer instead" #-}
+hamletToRepHtml = withUrlRenderer
+{-# DEPRECATED hamletToRepHtml "Use withUrlRenderer instead" #-}
 
--- | Provide a URL rendering function to the given function and return the
--- result. Useful for processing Shakespearean templates.
+-- | Deprecated synonym for 'withUrlRenderer'.
 --
 -- Since 1.2.0
 giveUrlRenderer :: MonadHandler m
                 => ((Route (HandlerSite m) -> [(Text, Text)] -> Text) -> output)
                 -> m output
-giveUrlRenderer f = do
+giveUrlRenderer = withUrlRenderer
+{-# DEPRECATED giveUrlRenderer "Use withUrlRenderer instead" #-}
+
+-- | Provide a URL rendering function to the given function and return the
+-- result. Useful for processing Shakespearean templates.
+--
+-- Since 1.2.20
+withUrlRenderer :: MonadHandler m
+                => ((Route (HandlerSite m) -> [(Text, Text)] -> Text) -> output)
+                -> m output
+withUrlRenderer f = do
     render <- getUrlRenderParams
     return $ f render
 
@@ -895,34 +898,47 @@ getMessageRender = do
     l <- reqLangs `liftM` getRequest
     return $ renderMessage (rheSite env) l
 
--- | Use a per-request cache to avoid performing the same action multiple
--- times. Note that values are stored by their type. Therefore, you should use
--- newtype wrappers to distinguish logically different types.
+-- | Use a per-request cache to avoid performing the same action multiple times.
+-- Values are stored by their type, the result of typeOf from Typeable.
+-- Therefore, you should use different newtype wrappers at each cache site.
+--
+-- For example, yesod-auth uses an un-exported newtype, CachedMaybeAuth and exports functions that utilize it such as maybeAuth.
+-- This means that another module can create its own newtype wrapper to cache the same type from a different action without any cache conflicts.
+--
+-- See the original announcement: <http://www.yesodweb.com/blog/2013/03/yesod-1-2-cleaner-internals>
 --
 -- Since 1.2.0
 cached :: (MonadHandler m, Typeable a)
        => m a
        -> m a
-cached f = do
+cached action = do
     gs <- get
-    let cache = ghsCache gs
-    case clookup cache of
-        Just val -> return val
-        Nothing -> do
-            val <- f
-            put $ gs { ghsCache = cinsert val cache }
-            return val
-  where
-    clookup :: Typeable a => Cache -> Maybe a
-    clookup (Cache m) =
-        res
-      where
-        res = Map.lookup (typeOf $ fromJust res) m >>= fromDynamic
-        fromJust :: Maybe a -> a
-        fromJust = error "Yesod.Handler.cached.fromJust: Argument to typeOf was evaluated"
+    eres <- Cache.cached (ghsCache gs) action
+    case eres of
+      Right res -> return res
+      Left (newCache, res) -> do
+          put $ gs { ghsCache = newCache }
+          return res
 
-    cinsert :: Typeable a => a -> Cache -> Cache
-    cinsert v (Cache m) = Cache (Map.insert (typeOf v) (toDyn v) m)
+-- | a per-request cache. just like 'cached'.
+-- 'cached' can only cache a single value per type.
+-- 'cachedBy' stores multiple values per type by usage of a ByteString key
+--
+-- 'cached' is ideal to cache an action that has only one value of a type, such as the session's current user
+-- 'cachedBy' is required if the action has parameters and can return multiple values per type.
+-- You can turn those parameters into a ByteString cache key.
+-- For example, caching a lookup of a Link by a token where multiple token lookups might be performed.
+--
+-- Since 1.4.0
+cachedBy :: (MonadHandler m, Typeable a) => S.ByteString -> m a -> m a
+cachedBy k action = do
+    gs <- get
+    eres <- Cache.cachedBy (ghsCacheBy gs) k action
+    case eres of
+      Right res -> return res
+      Left (newCache, res) -> do
+          put $ gs { ghsCacheBy = newCache }
+          return res
 
 -- | Get the list of supported languages supplied by the user.
 --
@@ -961,6 +977,30 @@ lookupHeaders key = do
     req <- waiRequest
     return $ lookup' key $ W.requestHeaders req
 
+-- | Lookup basic authentication data from __Authorization__ header of
+-- request. Returns user name and password
+--
+-- Since 1.4.9
+lookupBasicAuth :: (MonadHandler m) => m (Maybe (Text, Text))
+lookupBasicAuth = fmap (>>= getBA)
+                  (lookupHeader "Authorization")
+  where
+    getBA bs = (\(x, y) -> ( decodeUtf8With lenientDecode x
+                          , decodeUtf8With lenientDecode y))
+               <$> extractBasicAuth bs
+
+-- | Lookup bearer authentication datafrom __Authorization__ header of
+-- request. Returns bearer token value
+--
+-- Since 1.4.9
+lookupBearerAuth :: (MonadHandler m) => m (Maybe Text)
+lookupBearerAuth = fmap (>>= getBR)
+                   (lookupHeader "Authorization")
+  where
+    getBR bs = decodeUtf8With lenientDecode
+               <$> extractBearerAuth bs
+
+
 -- | Lookup for GET parameters.
 lookupGetParams :: MonadHandler m => Text -> m [Text]
 lookupGetParams pn = do
@@ -1116,22 +1156,12 @@ provideRepType ct handler =
 rawRequestBody :: MonadHandler m => Source m S.ByteString
 rawRequestBody = do
     req <- lift waiRequest
-#if MIN_VERSION_wai(3, 0, 0)
     let loop = do
             bs <- liftIO $ W.requestBody req
             unless (S.null bs) $ do
                 yield bs
                 loop
     loop
-#else
-    transPipe
-#if MIN_VERSION_wai(2, 0, 0)
-        liftIO
-#else
-        liftResourceT
-#endif
-        (W.requestBody req)
-#endif
 
 -- | Stream the data from the file. Since Yesod 1.2, this has been generalized
 -- to work in any @MonadResource@.
