@@ -10,7 +10,7 @@ module Yesod.Core.Internal.Run where
 import Yesod.Core.Internal.Response
 import           Blaze.ByteString.Builder     (toByteString)
 import           Control.Applicative          ((<$>))
-import           Control.Exception            (fromException, bracketOnError, evaluate)
+import           Control.Exception            (fromException, evaluate)
 import qualified Control.Exception            as E
 import           Control.Exception.Lifted     (catch)
 import           Control.Monad                (mplus)
@@ -18,7 +18,7 @@ import           Control.Monad.IO.Class       (MonadIO)
 import           Control.Monad.IO.Class       (liftIO)
 import           Control.Monad.Logger         (LogLevel (LevelError), LogSource,
                                                liftLoc)
-import           Control.Monad.Trans.Resource (runResourceT, withInternalState, runInternalState, createInternalState, closeInternalState)
+import           Control.Monad.Trans.Resource (runResourceT, withInternalState, runInternalState)
 import qualified Data.ByteString              as S
 import qualified Data.ByteString.Char8        as S8
 import qualified Data.IORef                   as I
@@ -31,15 +31,13 @@ import qualified Data.Text                    as T
 import           Data.Text.Encoding           (encodeUtf8)
 import           Data.Text.Encoding           (decodeUtf8With)
 import           Data.Text.Encoding.Error     (lenientDecode)
+import           Data.Time                    (getCurrentTime, addUTCTime)
 import           Language.Haskell.TH.Syntax   (Loc, qLocation)
 import qualified Network.HTTP.Types           as H
 import           Network.Wai
-#if MIN_VERSION_wai(2, 0, 0)
 import           Network.Wai.Internal
-#endif
+#if !MIN_VERSION_base(4, 6, 0)
 import           Prelude                      hiding (catch)
-#if !MIN_VERSION_fast_logger(2, 0, 0)
-import           System.Log.FastLogger        (Logger)
 #endif
 import           System.Log.FastLogger        (LogStr, toLogStr)
 import           System.Random                (newStdGen)
@@ -48,9 +46,11 @@ import           Yesod.Core.Class.Yesod
 import           Yesod.Core.Types
 import           Yesod.Core.Internal.Request  (parseWaiRequest,
                                                tooLargeResponse)
+import           Yesod.Core.Internal.Util     (formatRFC1123)
 import           Yesod.Routes.Class           (Route, renderRoute)
 import Control.DeepSeq (($!!), NFData)
 import Control.Monad (liftM)
+import Control.AutoUpdate (mkAutoUpdate, defaultUpdateSettings, updateAction, updateFreq)
 
 returnDeepSessionMap :: Monad m => SessionMap -> m SessionMap
 #if MIN_VERSION_bytestring(0, 10, 0)
@@ -79,6 +79,7 @@ runHandler rhe@RunHandlerEnv {..} handler yreq = withInternalState $ \resState -
         , ghsRBC = Nothing
         , ghsIdent = 1
         , ghsCache = mempty
+        , ghsCacheBy = mempty
         , ghsHeaders = mempty
         }
     let hd = HandlerData
@@ -119,6 +120,7 @@ runHandler rhe@RunHandlerEnv {..} handler yreq = withInternalState $ \resState -
                             | otherwise = status'
                      in return $ YRPlain status hs' ct c sess
                 YRWai _ -> return yar
+                YRWaiApp _ -> return yar
     let sendFile' ct fp p =
             return $ YRPlain H.status200 headers ct (ContentFile fp p) finalSession
     contents1 <- evaluate contents `E.catch` \e -> return
@@ -195,6 +197,7 @@ runFakeHandler :: (Yesod site, MonadIO m) =>
                -> m (Either ErrorResponse a)
 runFakeHandler fakeSessionMap logger site handler = liftIO $ do
   ret <- I.newIORef (Left $ InternalError "runFakeHandler: no result")
+  getMaxExpires <- getGetMaxExpires
   let handler' = do liftIO . I.writeIORef ret . Right =<< handler
                     return ()
   let yapp = runHandler
@@ -205,6 +208,7 @@ runFakeHandler fakeSessionMap logger site handler = liftIO $ do
             , rheUpload = fileUpload site
             , rheLog = messageLoggerSource site $ logger site
             , rheOnError = errHandler
+            , rheGetMaxExpires = getMaxExpires
             }
         handler'
       errHandler err req = do
@@ -220,24 +224,16 @@ runFakeHandler fakeSessionMap logger site handler = liftIO $ do
           , httpVersion    = H.http11
           , rawPathInfo    = "/runFakeHandler/pathInfo"
           , rawQueryString = ""
-#if MIN_VERSION_wai(2, 0, 0)
           , requestHeaderHost = Nothing
-#else
-          , serverName     = "runFakeHandler-serverName"
-          , serverPort     = 80
-#endif
           , requestHeaders = []
           , isSecure       = False
           , remoteHost     = error "runFakeHandler-remoteHost"
           , pathInfo       = ["runFakeHandler", "pathInfo"]
           , queryString    = []
-#if MIN_VERSION_wai(3, 0, 0)
           , requestBody    = return mempty
-#else
-          , requestBody    = mempty
-#endif
           , vault          = mempty
           , requestBodyLength = KnownLength 0
+          , requestHeaderRange = Nothing
           }
       fakeRequest =
         YesodRequest
@@ -251,29 +247,24 @@ runFakeHandler fakeSessionMap logger site handler = liftIO $ do
           }
   _ <- runResourceT $ yapp fakeRequest
   I.readIORef ret
-{-# WARNING runFakeHandler "Usually you should *not* use runFakeHandler unless you really understand how it works and why you need it." #-}
 
 yesodRunner :: (ToTypedContent res, Yesod site)
             => HandlerT site IO res
             -> YesodRunnerEnv site
             -> Maybe (Route site)
             -> Application
-#if MIN_VERSION_wai(3, 0, 0)
 yesodRunner handler' YesodRunnerEnv {..} route req sendResponse
   | Just maxLen <- mmaxLen, KnownLength len <- requestBodyLength req, maxLen < len = sendResponse tooLargeResponse
-#else
-yesodRunner handler' YesodRunnerEnv {..} route req
-  | Just maxLen <- mmaxLen, KnownLength len <- requestBodyLength req, maxLen < len = return tooLargeResponse
-#endif
   | otherwise = do
     let dontSaveSession _ = return []
     (session, saveSession) <- liftIO $ do
         maybe (return (Map.empty, dontSaveSession)) (\sb -> sbLoadSession sb req) yreSessionBackend
+    getMaxExpires <- getGetMaxExpires
     let mkYesodReq = parseWaiRequest req session (isJust yreSessionBackend) mmaxLen
-    yreq <-
-        case mkYesodReq of
-            Left yreq -> return yreq
-            Right needGen -> liftIO $ needGen <$> newStdGen
+    let yreq =
+            case mkYesodReq of
+                Left yreq -> yreq
+                Right needGen -> needGen yreGen
     let ra = resolveApproot yreSite req
     let log' = messageLoggerSource yreSite yreLogger
         -- We set up two environments: the first one has a "safe" error handler
@@ -287,33 +278,26 @@ yesodRunner handler' YesodRunnerEnv {..} route req
             , rheUpload = fileUpload yreSite
             , rheLog = log'
             , rheOnError = safeEh log'
+            , rheGetMaxExpires = getMaxExpires
             }
         rhe = rheSafe
             { rheOnError = runHandler rheSafe . errorHandler
             }
-#if MIN_VERSION_wai(3, 0, 0)
 
-    E.bracket createInternalState closeInternalState $ \is -> do
+    yesodWithInternalState yreSite route $ \is -> do
         yreq' <- yreq
         yar <- runInternalState (runHandler rhe handler yreq') is
         yarToResponse yar saveSession yreq' req is sendResponse
-
-#else
-
-#if MIN_VERSION_wai(2, 0, 0)
-    bracketOnError createInternalState closeInternalState $ \is -> do
-        yreq' <- yreq
-        yar <- runInternalState (runHandler rhe handler yreq') is
-        liftIO $ yarToResponse yar saveSession yreq' req is
-#else
-    yar <- runHandler rhe handler yreq
-    liftIO $ yarToResponse yar saveSession yreq req
-#endif
-#endif
   where
     mmaxLen = maximumContentLength yreSite route
     handler = yesodMiddleware handler'
 
+getGetMaxExpires :: MonadIO m => m (IO Text)
+getGetMaxExpires = liftIO $ mkAutoUpdate defaultUpdateSettings
+  { updateAction = liftM (formatRFC1123 . addUTCTime (60*60*24*365)) getCurrentTime
+  , updateFreq = 60 * 60 * 1000000 -- Update once per hour
+  }
+
 yesodRender :: Yesod y
             => y
             -> ResolvedApproot
